@@ -1,7 +1,7 @@
 package Bencher;
 
-our $DATE = '2015-12-23'; # DATE
-our $VERSION = '0.19'; # VERSION
+our $DATE = '2016-01-05'; # DATE
+our $VERSION = '0.20'; # VERSION
 
 use 5.010001;
 use strict;
@@ -545,7 +545,7 @@ sub _gen_items {
                 };
             } elsif ($p->{type} eq 'perl_code') {
                 my $args;
-                if ($ds->{args}) {
+                if ($ds && $ds->{args}) {
                     $args = { %{$ds->{args}} };
                     delete $args->{$_} for (grep {/\@\z/} keys %$args);
                     $args->{$_} = $h_args->{$_} for keys %$h_args;
@@ -862,6 +862,72 @@ my $_alias_spec_add_participant = {
     },
 };
 
+$SPEC{format_result} = {
+    v => 1.1,
+    summary => 'Format bencher result',
+    args => {
+        envres => {
+            summary => 'Enveloped result from bencher',
+            schema => 'array*', # XXX envres
+            req => 1,
+            pos => 0,
+            cmdline_src => 'stdin_or_file',
+        },
+    },
+    args_as => 'array',
+};
+sub format_result {
+    my $envres = shift;
+
+    my $ff = $envres->[3]{'table.fields'};
+
+    # for module_startup mode: remove 'rate', add 'mod_overhead_time'
+    if ($envres->[3]{'func.module_startup'}) {
+        my $rit_baseline = _find_record_by_seq($envres->[2], 0);
+        for my $rit (@{$envres->[2]}) {
+            delete $rit->{rate};
+            if ($rit_baseline) {
+                $rit->{mod_overhead_time} = $rit->{time} - $rit_baseline->{time};
+            }
+        }
+        splice @$ff, (firstidx {$_ eq 'rate'} @$ff), 1;
+        splice @$ff, (firstidx {$_ eq 'time'} @$ff)+1, 0, "mod_overhead_time";
+    }
+
+    # sort by default from slowest to fastest
+    $envres->[2] = [sort {$b->{time} <=> $a->{time}} @{$envres->[2]}];
+
+    # pick an appropriate time unit & format the time
+    my ($min, $max) = minmax(map {$_->{time}} @{$envres->[2]});
+
+    my ($unit, $factor);
+    if ($max <= 1.5e-6) {
+        $unit = "ns";
+        $factor = 1e9;
+    } elsif ($max <= 1.5e-3) {
+        $unit = "\x{03bc}s"; # XXX enable utf
+        $factor = 1e6;
+    } elsif ($max <= 1.5) {
+        $unit = "ms";
+        $factor = 1e3;
+    }
+
+    if ($unit) {
+        for my $rit (@{$envres->[2]}) {
+            # XXX format number of decimal digits of 'time' and 'rate'
+            # based on sigma
+            $rit->{time} = sprintf(
+                "%.5f%s", $rit->{time} * $factor, $unit);
+            if (exists $rit->{mod_overhead_time}) {
+                $rit->{mod_overhead_time} = sprintf(
+                    "%.5f%s", $rit->{mod_overhead_time} * $factor, $unit);
+            }
+        }
+    }
+
+    $envres;
+}
+
 $SPEC{bencher} = {
     v => 1.1,
     summary => 'A benchmark framework',
@@ -1363,6 +1429,7 @@ sub bencher {
             $args{-cmdline_r} && (($args{-cmdline_r}{format} // '') !~ /json/) ?
             0 : 1;
 
+        $envres->[3]{'func.module_startup'} = $module_startup;
         $envres->[3]{'func.module_versions'}{perl} = $^V if $return_resmeta;
 
         my $code_load = sub {
@@ -1455,6 +1522,18 @@ sub bencher {
         }
 
         if ($return_resmeta) {
+            $envres->[3]{'func.bencher_version'} = $Bencher::VERSION;
+            $envres->[3]{'func.bencher_args'} = {
+                map {$_=>$args{$_}} grep {!/\A-/} keys %args};
+            if ($args{scenario_file}) {
+                $envres->[3]{'func.scenario_file'} = $args{scenario_file};
+            } elsif (my $mod = $args{scenario_module}) {
+                $mod = "Bencher::Scenario::$mod";
+                no strict 'refs';
+                $envres->[3]{'func.scenario_module'} = $mod;
+                $envres->[3]{'func.module_versions'}{$mod} =
+                    ${"$mod\::VERSION"};
+            }
             $envres->[3]{'func.sysload_before'} = [Sys::Load::getload()];
             $envres->[3]{'_time_start'} = time();
         }
@@ -1503,53 +1582,8 @@ sub bencher {
             my $r = $args{-cmdline_r};
             last unless $r && ($r->{format} // 'text') =~ /text/;
 
-            my $ff = $envres->[3]{'table.fields'};
-
-            # for module_startup mode: remove 'rate', add 'mod_overhead_time'
-            if ($module_startup) {
-                my $rit_baseline = _find_record_by_seq($envres->[2], 0);
-                for my $rit (@{$envres->[2]}) {
-                    delete $rit->{rate};
-                    if ($rit_baseline) {
-                        $rit->{mod_overhead_time} = $rit->{time} - $rit_baseline->{time};
-                    }
-                }
-                splice @$ff, (firstidx {$_ eq 'rate'} @$ff), 1;
-                splice @$ff, (firstidx {$_ eq 'time'} @$ff)+1, 0, "mod_overhead_time";
-            }
-
-            # sort by default from slowest to fastest
-            $envres->[2] = [sort {$b->{time} <=> $a->{time}} @{$envres->[2]}];
-
-            # pick an appropriate time unit & format the time
-            my ($min, $max) = minmax(map {$_->{time}} @{$envres->[2]});
-
-            my ($unit, $factor);
-            if ($max <= 1.5e-6) {
-                $unit = "ns";
-                $factor = 1e9;
-            } elsif ($max <= 1.5e-3) {
-                $unit = "\x{03bc}s"; # XXX enable utf
-                $factor = 1e6;
-            } elsif ($max <= 1.5) {
-                $unit = "ms";
-                $factor = 1e3;
-            }
-
-            if ($unit) {
-                for my $rit (@{$envres->[2]}) {
-                    # XXX format number of decimal digits of 'time' and 'rate'
-                    # based on sigma
-                    $rit->{time} = sprintf(
-                        "%.5f%s", $rit->{time} * $factor, $unit);
-                    if (exists $rit->{mod_overhead_time}) {
-                        $rit->{mod_overhead_time} = sprintf(
-                            "%.5f%s", $rit->{mod_overhead_time} * $factor, $unit);
-                    }
-                }
-            }
-
-        } # FORMAT
+            format_result($envres);
+        }
 
         if ($return_resmeta) {
             $envres->[3]{'func.platform_info'} =
@@ -1612,7 +1646,7 @@ Bencher - A benchmark framework
 
 =head1 VERSION
 
-This document describes version 0.19 of Bencher (from Perl distribution Bencher), released on 2015-12-23.
+This document describes version 0.20 of Bencher (from Perl distribution Bencher), released on 2016-01-05.
 
 =head1 SYNOPSIS
 
@@ -1821,6 +1855,34 @@ that contains extra information.
 Return value:  (any)
 
 
+=head2 format_result($envres) -> [status, msg, result, meta]
+
+Format bencher result.
+
+This function is not exportable.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<envres>* => I<array>
+
+Enveloped result from bencher.
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
+
+
 =head2 parse_scenario(%args) -> [status, msg, result, meta]
 
 Parse scenario (fill in default values, etc).
@@ -1902,9 +1964,22 @@ version will use shell. See L<Bencher::Scenario::Interpreters>.
 
 =over
 
-=item * name
+=item * name (str)
 
-=item * summary
+From DefHash.
+
+=item * summary (str)
+
+From DefHash.
+
+=item * description (str)
+
+From DefHash.
+
+=item * tags (array of str)
+
+From DefHash. Define tag(s) for this participant. Can be used to include/exclude
+groups of participants having the same tags.
 
 =item * module (str)
 
@@ -1927,13 +2002,24 @@ benchmark, unless the dataset is explicitly included).
 
 =over
 
-=item * name
+=item * name (str)
 
-=item * summary
+From DefHash.
 
-=item * description
+=item * summary (str)
 
-=item * args
+From DefHash.
+
+=item * description (str)
+
+From DefHash.
+
+=item * tags (array of str)
+
+From DefHash. Define tag(s) for this dataset. Can be used to include/exclude
+groups of datasets having the same tags.
+
+=item * args (hash)
 
 Example:
 
@@ -1947,10 +2033,18 @@ name, example:
 This means, for each participant mentioning C<size>, three benchmark items will
 be generated, one for each value of C<size>.
 
-=item * argv
+=item * argv (array)
 
 =item * include_by_default (bool, default 1)
 
+=item * include_participant_tags (array of str)
+
+Only include participants having all these tags.
+
+=item * exclude_participant_tags (array of str)
+
+Exclude participants having any of these tags.
+
 =back
 
 =head3 Other properties
@@ -2044,7 +2138,7 @@ perlancar <perlancar@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2015 by perlancar@cpan.org.
+This software is copyright (c) 2016 by perlancar@cpan.org.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
